#!/usr/bin/env python3

import sys, os

#TODO if payload[b'\x45\x4e\x46'] exists then concatenate the payload with value at 0x454E45 to prevent getting an EOF
#TODO check if file is a patch and asks if user wants to apply to patch instead
#TODO create -c option for the user to specify the desire to create a patch
#TODO manage SNES headers

def main(args):
    print('IPS patch creator tool v0.5')
    print('===========================\n')
    if len(args) == 2:
        create_patch(args)
    else:
        print('Please enter the name of the source and destination files like this (without the quotes): "python3',__file__,'source_file destination_file"')
        #menu_choice = "What do you want to do? ('Q' to exit)\n" + "\t1. Create a IPS patch" + "\n\n===>"
        #choice = input(menu_choice)
        #while (choice.lower() != 'q'):
        #    if choice == '1':
        #        create_patch(None)
        #    else:
        #        print('Your choice is invalid. Please try again')
        #    input(menu_choice)

def create_patch(args):
    #if args is None:
    #    source_filename = input('Path of the source ROM: ')
    #    destination_filename = input('Path of the destination ROM: ')
    #else:
    source_filename = args[0]
    destination_filename = args[1]
    try:
        source_size = os.stat(source_filename).st_size
        source_file = open(source_filename, 'rb')
        destination_size = os.stat(destination_filename).st_size
        destination_file = open(destination_filename, 'rb')
    except FileNotFoundError as err:
        print(err)

    print('Begin of the analysis of the files')
    payloads = {}
    #Checks which file is smaller
    patch_range = source_size if source_size < destination_size else destination_size
    #The IPS format can only changes the first 2^24 values
    patch_range = patch_range if patch_range < 16777216 else 16777216
    for i in range(patch_range):
        source_byte = source_file.read(1)
        destination_byte = destination_file.read(1)
        if source_byte != destination_byte:
            payloads[i] = destination_byte
    source_file.close()
    destination_file.close()

    print('Analysis done! Beginning of the compression')

    payloads = compact_payloads(payloads)

    patch_filename = input('Compression done! IPS patch is ready to be created.\nPlease insert the filename for the new patch (default is "patch.ips"): ')
    patch_filename = patch_filename if patch_filename != '' else 'patch.ips'
    create_patch_file(patch_filename, payloads)

    print('Patch created. Enjoy!')
    
def create_patch_file(filename, payloads):
    try:
        patch_file = open(filename, 'wb')
        patch_file.write(b'PATCH')
        for address, payload in sorted(payloads.items()):
            write_payload(patch_file, address, payload)
        patch_file.write(b'EOF')
        patch_file.close()
    except Error as err:
        print(err)


#This function takes a stream and write the patch in it.
def write_payload(file, address, payload):
    file.write(address.to_bytes(3, 'big'))
    if isinstance(payload, list):
        file.write(b'\x00\x00')
        file.write(payload[0].to_bytes(2, 'big'))
        file.write(payload[1].to_bytes(1, 'big'))
    else:
        file.write(len(payload).to_bytes(2, 'big'))
        file.write(payload)

def compact_payloads(payloads):
    reverse_list = list(reversed(sorted(payloads.keys())))
    #Iterates over the keys of the payloads in reverse order so it is easier to concatenate entries, 
    for i in reverse_list:
        #A payload length is 2 bytes, so the size must be smaller than FFFFh
        if i-1 in payloads and sys.getsizeof(payloads[i]) < 65536:
            #concatenate current entry with previous entry and delete current entry
            payloads[i-1] += payloads[i]
            del payloads[i]

    #Creates a duplicate payloads to work on in order to not disturb the loop size
    rle_payloads = dict(payloads)

    #Splits the entries that contains a treak of at least 14 of the same character in a row
    for key, value in payloads.items():

        #Do not bother to do this if the payload is too short
        if len(value) > 13:
            for i in range(len(value) - 13):
                j = i + 1
                while (j < len(value) and value[i] == value[j]):
                    j += 1
                #If the streak of the same character is long enough, then split into new entries accordingly.
                if j - i > 13:
                    #If the streak happen in the middle of the payload, then update the initial entry. If not, delete the entry.
                    if i > 0:
                        rle_payloads[key] = value[:i]
                    else:
                        del rle_payloads[key]
                    #Creates a payload for the streak itself.
                    #By changing the byte sequence into a list, it becomes easier later to determine whether or not it is a RLE compression.
                    rle_payloads[key+i] = [j-i, value[i]]
                    #If the streak does not happen at the end of the payload, then create a new entry for the rest of the payload.
                    if j < len(value):
                        rle_payloads[key+j] = value[j:]
                    #If there is another streak in the same payload, we need to update which payload will actually be splitted.
                    key += j
                    i = j

        #There are two situations where this is useful:
        #
        #1 - If the payload is one whole streak of 4 to 13 characters
        #2 - If, by splitting the streaks in a very long payload, the last one happens to be a streak between 4 and 13 characters.
        if key in rle_payloads and len(rle_payloads[key]) > 4 and len(rle_payloads[key]) <= 13:
            value = rle_payloads[key]
            streak = True
            i = 1
            while (i < len(value) and streak == True):
                streak = True if value[0] == value[i] else False
            if streak:
                rle_payloads[key] = [len(value), value[0]]
                
            
    return rle_payloads

main(sys.argv[1:])























